from collections import defaultdict
from row import Row, StreamInfo, PointStream
from device import Device
from sensor import Sensor


def make_row_generator(rows):
    """"Utility function for converting a list to a generator.

    :param list d: the list to convert
    :rtype: generator"""

    for r in rows:
        yield Row(r)


def make_device_generator(devices):
    for d in devices:
        sensors = []
        for s in d['sensors']:
            sensors.append(Sensor(s['key'], s.get('name', ''),
                                  s['attributes']))
        yield Device(d['key'], d.get('name', ''), d['attributes'], sensors)


def check_response(resp):
    """Utility function for checking the status of a cursor increment.  Raises
    an exception if the call to the paginated link returns anything other than
    a 200.

    :param resp: the response to check
    :type resp: :class:`tempodb.response.Response` object
    :raises ValueError: if the response is not a 200
    :rtype: None"""

    s = resp.status
    if s != 200:
        raise ValueError(
            'TempoIQ API returned %d as status when 200 was expected' % s)


class Cursor(object):
    """An iterable cursor over data retrieved from the TempoDB API.  The
    cursor will make network requests to fetch more data as needed, until
    the API returns no more data.  It can be used with the standard
    iterable interface::

        >>> data = [d for d in response.data]"""

    def __init__(self, data,  response):
        self.response = response
        self.data = make_row_generator(data)

    def __iter__(self):
        while True:
            try:
                x = self.data.next()
                yield x
            except StopIteration:
                self._fetch_next()

    def _fetch_next(self):
        raise StopIteration


class DeviceCursor(Cursor):
    """The data attribute holds the actual data from the request.

    Additionally, the raw response object is available as the response
    attribute of the cursor.

    :param response: the raw response object
    :type response: :class:`tempodb.response.Response"""

    def __init__(self, response, data, fetcher):
        self.response = response
        self.fetcher = fetcher
        self._raw_data = data
        self.data = make_device_generator(data['data'])

    def _fetch_next(self):
        try:
            cursor_obj = self._raw_data['next_page']['next_query']
            new_data = self.fetcher(cursor_obj)
            self._raw_data = new_data
            self.response.data = new_data
            self.data = make_device_generator(new_data['data'])
        except KeyError:
            raise StopIteration


class DataPointsCursor(Cursor):
    """The data attribute holds the actual data from the request.

    Additionally, the raw response object is available as the response
    attribute of the cursor.

    :param response: the raw response object
    :type response: :class:`tempodb.response.Response"""

    def __init__(self, response, data, fetcher):
        self.response = response
        self.fetcher = fetcher
        self._raw_data = data
        self.data = make_row_generator(data['data'])

    def _fetch_next(self):
        try:
            cursor_obj = self._raw_data['next_page']['next_query']
            new_data = self.fetcher(cursor_obj)
            self._raw_data = new_data
            self.response.data = new_data
            self.data = make_row_generator(new_data['data'])
        except KeyError:
            raise StopIteration
        except Exception:
            raise


class Page(object):
    def __init__(self, data, cursor_obj, collectible=True):
        self.data = data
        self.cursor_obj = cursor_obj
        self.is_collectible = collectible

    def garbage_collect(self):
        self.data = None

    def is_active(self):
        if self.data is not None:
            return True
        return False

    def reconstruct(self, cursor):
        try:
            self.data = cursor.fetcher(self.cursor_obj)['data']
        except Exception:
            raise


class StreamManager(object):
    MAX_PAGES = 100

    def __init__(self, cursor, starting_data, page_size):
        if starting_data.get('next_page') is not None:
            cursor_obj = starting_data['next_page']['next_query']
        else:
            cursor_obj = None
        self.pages = {0: Page(starting_data['data'], None, False)}
        self.active_pages = 1
        self.receiver_pointers = {}
        self.active_pointers = defaultdict(set)
        self.cursor = cursor
        self.page_size = page_size

    def _garbage_collect(self):
        to_collect = self.active_pages - StreamManager.MAX_PAGES
        max_iters = 2
        current_iters = 0
        while to_collect > 0 and current_iters < max_iters:
            for k in self.pages:
                if len(self.active_pointers[k]) == 0:
                    page = self.pages[k]
                    if page.is_collectible:
                        page.garbage_collect()
                        self.active_pages -= 1
                        to_collect -= 1
                        if to_collect == 0:
                            break
                    else:
                        continue
                    #dont want to be too aggressive here, we are guaranteeing
                    #max memory used and trying to keep network traffic
                    #to a minimum
            current_iters += 1

    def _update_active_pointers(self, key, page_num):
        if page_num == 0:
            self.active_pointers[page_num].add(key)
        elif key not in self.active_pointers[page_num]:
            old_page = page_num - 1
            self.active_pointers[old_page].remove(key)
            self.active_pointers[page_num].add(key)

    def fetch_next(self):
        next_idx = self.pages.keys()[-1] + 1
        new_data, cursor_obj = self.cursor._fetch_next()
        page = Page(new_data['data'], cursor_obj)
        self.pages[next_idx] = page
        self.active_pages += 1
        return page

    def next(self, receiver):
        if not self.receiver_pointers.get(receiver.key):
            self.receiver_pointers[receiver.key] = 0
            self.active_pointers[0].add(receiver.key)

        current_idx = self.receiver_pointers[receiver.key]
        page_num = current_idx / self.page_size
        item = current_idx % self.page_size
        self.receiver_pointers[receiver.key] = current_idx + 1
        self._update_active_pointers(receiver.key, page_num)

        page = self.pages.get(page_num)
        if page is None:
            page = self.fetch_next()
            self._garbage_collect()
        if not page.is_active():
            self.reconstruct_page(page)
        return page.data[item]

    def reconstruct_page(self, page):
        page.reconstruct(self.cursor)
        self.active_pages += 1


class StreamResponseCursor(Cursor):
    def __init__(self, response, data, fetcher):
        self.response = response
        self.fetcher = fetcher
        self._raw_data = data
        self.page_size = len(data['data'])
        self.stream_info = StreamInfo(data['streams'])
        self.manager = StreamManager(self, data, self.page_size)

    def __iter__(self):
        streams = self.streams
        iterators = [(s, iter(s)) for s in streams]
        for stream, it in iterators:
            yield ((stream.device, stream.sensor), it.next())

    def _fetch_next(self):
        try:
            cursor_obj = self._raw_data['next_page']['next_query']
            new_data = self.fetcher(cursor_obj)
            self._raw_data = new_data
            if len(new_data['data']) == 0:
                raise StopIteration
            return new_data, cursor_obj
        except KeyError:
            raise StopIteration
        except Exception:
            raise

    def bind_stream(self, **kwargs):
        stream_info = self.stream_info.get_one(**kwargs)
        return PointStream(stream_info, self.manager)

    @property
    def streams(self):
        return [PointStream(si, self.manager) for
                si in self.stream_info.headers]



class Compound(object):
    def __init__(self):
        self.selectors = []
        self.selection_type = None

    def add(self, selector):
        self.selectors.append(selector)


class AndClause(Compound):
    pass


class OrClause(Compound):
    pass


class ScalarSelector(object):
    """A condition to filter on. The recommended way to construct a ScalarSelector
    is with the ``==`` operator on a selectable field, e.g::

        selector = (Device.attributes['model'] == "v1")
        selector2 = (Sensor.key == "humidity")
    """
    def __init__(self, selection_type, key, value):
        self.selection_type = selection_type
        self.key = key
        self.value = value


class Selectable(object):
    def __init__(self, selection_type, key):
        self.selection_type = selection_type
        self.key = key


class ScalarSelectable(Selectable):
    def __init__(self, selection_type, key):
        self.selection_type = selection_type
        self.key = key

    def __eq__(self, other):
        return ScalarSelector(self.selection_type, self.key, other)


class ItemProxy(object):
    def __init__(self, selection_type, name, key):
        self.selection_type = selection_type
        self.name = name
        self.key = key

    def __eq__(self, other):
        return ScalarSelector(self.selection_type, self.name,
                              {self.key: other})


class DictSelectable(Selectable):
    def __init__(self, selection_type, key):
        self.selection_type = selection_type
        self.key = key

    def __getitem__(self, key):
        return ItemProxy(self.selection_type, self.key, key)


class Selection(object):
    def __init__(self):
        self.selection = None

    def add(self, selector):
        if self.selection is None:
            self.selection = selector
        else:
            if issubclass(selector.__class__, Compound):
                self.selection.add(selector)
            else:
                clause = AndClause()
                clause.add(self.selection)
                clause.add(selector)
                self.selection = clause


def and_(selectors):
    """Returns a selector that's the AND of all provided selectors.

    :param selectors:
    :type selectors: List of :class:`ScalarSelector`"""
    s = AndClause()
    object_type = None
    for selector in selectors:
        if object_type is None:
            object_type = selector.selection_type
        s.add(selector)
    s.selection_type = object_type
    return s


def or_(selectors):
    """Returns a selector that's the OR of all provided selectors.

    :param selectors:
    :type selectors: List of :class:`ScalarSelector`"""
    s = OrClause()
    object_type = None
    for selector in selectors:
        if object_type is None:
            object_type = selector.selection_type
        s.add(selector)
    s.selection_type = object_type
    return s



from query.selection import ScalarSelectable, DictSelectable


class Device(object):
    """Representation of a Device in TempoIQ.

    :param string key:
    :param string name:
    :param dict attributes:
    :param list sensors:
    """
    key = ScalarSelectable('devices', 'key')
    name = ScalarSelectable('devices', 'name')
    attributes = DictSelectable('devices', 'attributes')

    def __init__(self, key, name='', attributes={}, sensors=[]):
        self.key = key
        self.name = name
        self.attributes = attributes
        self.sensors = sensors



import copy
from query.selection import ScalarSelectable


class Rule(object):
    key = ScalarSelectable('rules', 'key')

    def __init__(self, name, alert_by=None, key=None, selection=None,
                 conditions=[], action=None):
        self.name = name
        self.alert_by = alert_by
        self.key = key
        self.selection = selection
        #make sure to avoid mutability of kwargs that default to list here
        self.conditions = copy.deepcopy(conditions)
        self.action = action


class Trigger(object):
    def __init__(self, trigger_type, args):
        self.trigger_type = trigger_type
        self.args = args


class Filter(object):
    def __init__(self, inclusion, filter_type, args):
        self.inclusion = inclusion
        self.filter_type = filter_type
        self.args = args


class Condition(object):
    def __init__(self, filters, trigger):
        self.filters = filters
        self.trigger = trigger


class Action(object):
    pass


class Webhook(Action):
    def __init__(self, url):
        self.url = url



import json
from protocol.cursor import DeviceCursor, StreamResponseCursor
from protocol.cursor import DataPointsCursor
from protocol.decoder import TempoIQDecoder


SUCCESS = 0
FAILURE = 1
PARTIAL = 2


class ResponseException(Exception):
    """Exception class for HTTP responses"""

    def __init__(self, response):
        self.response = response
        self.msg = 'TempoIQ response returned status: %d' % response.status

    def __repr__(self):
        return self.msg

    def __str__(self):
        return self.msg


class Response(object):
    """Represents responses from the TempoIQ API.  The Response object has
    several useful attributes after it is created:

        * successful: whether the overall request was successful (see below)
        * status: the HTTP status code for the API call
        * reason: the explanation for the HTTP status code
        * data: an object or list of objects representing the data from the API
        * error: a string if the API returned any additional error information
                 in the response body, None otherwise

    **Note:** data will be None if the status code was anything other than
    200.

    **Note:** successful has 3 possible values defined as constants in this
    module, SUCCESS, FAILURE, and PARTIAL.  A PARTIAL value can occur during a
    multi-write if some datapoints fail to write.  The error attribute in that
    case will be a JSON encoded string of errors for each datapoint.  The
    response object does *not* derserialize that error (there could be other
    circumstances where the error is not JSON encoded), so if that case, error
    handling code surrounding multi-writes should decode the error attribute
    with the json library if it wants to attempt error recovery.

    :param obj resp: a response object from the requests library"""

    def __init__(self, resp, session):
        self.resp = resp
        self.session = session
        self.status = resp.status_code
        self.reason = resp.reason
        if self.status == 200:
            self.successful = SUCCESS
            self.error = None
        elif self.status == 207:
            self.successful = PARTIAL
            self.error = self.resp.text
        else:
            self.successful = FAILURE
            self.error = self.resp.text

        self.resp.encoding = "UTF-8"
        self.body = self.resp.text
        self.data = None


class DeviceResponse(Response):
    def __init__(self, resp, session, fetcher):
        super(DeviceResponse, self).__init__(resp, session)
        self.fetcher = fetcher
        if self.successful == SUCCESS:
            self.parse(self.body)

    def parse(self, body):
        data = json.loads(body)
        self.data = DeviceCursor(self, data, self.fetcher)


class SensorPointsResponse(Response):
    def __init__(self, resp, session, fetcher):
        super(SensorPointsResponse, self).__init__(resp, session)
        self.fetcher = fetcher
        if self.successful == SUCCESS:
            self.parse(self.body)

    def parse(self, body):
        self.data = DataPointsCursor(self, json.loads(body), self.fetcher)


class StreamResponse(Response):
    def __init__(self, resp, session, fetcher):
        super(StreamResponse, self).__init__(resp, session)
        self.fetcher = fetcher
        if self.successful == SUCCESS:
            self.parse(self.body)

    def parse(self, body):
        self.data = StreamResponseCursor(self, json.loads(body), self.fetcher)


class RuleResponse(Response):
    def __init__(self, resp, session):
        super(RuleResponse, self).__init__(resp, session)
        if self.successful == SUCCESS:
            self.parse(self.body)

    def parse(self, body):
        decoder = TempoIQDecoder()
        self.data = json.loads(body, object_hook=decoder)


class DeleteDatapointsResponse(Response):
    def __init__(self, resp, session):
        super(DeleteDatapointsResponse, self).__init__(resp, session)
        self.parse(self.body)

    def parse(self, body):
        self.data = {}



from query.selection import ScalarSelectable, DictSelectable


class Sensor(object):
    """Representation of a Sensor in TempoIQ.

    :param string key:
    :param string name:
    :param dict attributes:
    """
    key = ScalarSelectable('sensors', 'key')
    name = ScalarSelectable('sensors', 'name')
    attributes = DictSelectable('sensors', 'attributes')

    def __init__(self, key, name='', attributes={}):
        self.key = key
        self.name = name
        self.attributes = attributes



import json
import urlparse
import urllib
from protocol.encoder import WriteEncoder, CreateEncoder, ReadEncoder
from protocol.query.builder import QueryBuilder
from response import Response, SensorPointsResponse, DeleteDatapointsResponse
from response import StreamResponse
from response import RuleResponse, DeviceResponse, ResponseException
from endpoint import media_type, media_types


def escape(s):
    return urllib.quote(s, safe='')


def make_fetcher(endpoint, url, headers={}):
    def fetcher(cursor):
        resp = endpoint.get(url, json.dumps(cursor), headers=headers)
        if resp.status_code != 200:
            #munge this so the ResponsException can work with it
            resp.status = resp.status_code
            raise ResponseException(resp)
        return json.loads(resp.text)
    return fetcher


class MonitoringClient(object):
    def __init__(self, endpoint):
        self.endpoint = endpoint

    def delete_rule(self, key):
        url = urlparse.urljoin(self.endpoint.base_url, 'monitors/' + key)
        resp = self.endpoint.delete(url)
        return Response(resp, self.endpoint)

    def get_annotations(self, key):
        url = urlparse.urljoin(self.endpoint.base_url,
                               'monitors/annotations/' + key)
        resp = self.endpoint.get(url)
        return Response(resp, self.endpoint)

    def get_changelog(self, key):
        url = urlparse.urljoin(self.endpoint.base_url,
                               'monitors/%s/changes/' % key)
        resp = self.endpoint.get(url)
        return Response(resp, self.endpoint)

    def get_logs(self, key):
        url = urlparse.urljoin(self.endpoint.base_url,
                               'monitors/%s/logs/' % key)
        resp = self.endpoint.get(url)
        return Response(resp, self.endpoint)

    def get_rule(self, key):
        url = urlparse.urljoin(self.endpoint.base_url, 'monitors/' + key)
        return RuleResponse(self.endpoint.get(url), self)

    def get_usage(self, key):
        url = urlparse.urljoin(self.endpoint.base_url,
                               'monitors/%s/usage/' % key)
        resp = self.endpoint.get(url)
        return Response(resp, self.endpoint)

    def list_rules(self):
        url = urlparse.urljoin(self.endpoint.base_url, 'monitors/')
        return RuleResponse(self.endpoint.get(url), self)


class Client(object):
    """Entry point for all TempoIQ API calls.

    :param endpoint: backend and credentials to connect to
    :type endpoint: tempoiq.endpoint.HTTPEndpoint
    """

    write_encoder = WriteEncoder()
    create_encoder = CreateEncoder()
    read_encoder = ReadEncoder()

    def __init__(self, endpoint, read_version='v2'):
        self.endpoint = endpoint
        self.monitoring_client = MonitoringClient(self.endpoint)
        self.DATAPOINT_ACCEPT_TYPE = media_type('datapoint-collection',
                                           read_version)
        self.ERROR_ACCEPT_TYPE = media_type('error', 'v1')
        self.DEVICE_ACCEPT_TYPE = media_type('device-collection', 'v2')
        self.QUERY_CONTENT_TYPE = media_type('query', 'v1')
        self.read_version = read_version

    def create_device(self, device):
        """Create a new device

        :param device:
        :type device: :class:`~tempoiq.protocol.device.Device`
        :rtype: :class:`tempoiq.response.Response` with a
                :class:`tempoiq.protocol.device.Device` data payload"""

        url = urlparse.urljoin(self.endpoint.base_url, 'devices/')
        j = json.dumps(device, default=self.create_encoder.default)
        resp = self.endpoint.post(url, j)
        return Response(resp, self.endpoint)

    def delete_device(self, query):
        """Delete devices that match the provided query.

        :param query:
        :type query: :class:`tempoiq.protocol.query.builder.QueryBuilder`
        :rtype: :class:`tempoiq.response.Response`"""

        url = urlparse.urljoin(self.endpoint.base_url, 'devices/')
        j = json.dumps(query, default=self.read_encoder.default)
        resp = self.endpoint.delete(url, j)
        return Response(resp, self.endpoint)

    def update_device(self, device):
        """Update the attributes of a device.

        :param device:
        :type device: :class:`~tempoiq.protocol.device.Device`
        :rtype: :class:`tempoiq.response.Response` with a
                :class:`tempoiq.protocol.device.Device` data payload"""

        path = '/'.join(['devices', escape(device.key)])
        url = urlparse.urljoin(self.endpoint.base_url, path)
        j = json.dumps(device, default=self.create_encoder.default)
        resp = self.endpoint.put(url, j)
        return Response(resp, self.endpoint)

    def delete_from_sensors(self, device_key, sensor_key, start, end):
        path = '/'.join(
            ['devices', escape(device_key), 'sensors',  escape(sensor_key),
             'datapoints'])
        url = urlparse.urljoin(self.endpoint.base_url, path)
        j = json.dumps({'start': start.isoformat(),
                        'stop': end.isoformat()})
        resp = self.endpoint.delete(url, j)
        return DeleteDatapointsResponse(resp, self.endpoint)

    def monitor(self, rule):
        url = urlparse.urljoin(self.endpoint.base_url, 'monitors/')
        rule_json = json.dumps(rule, default=self.write_encoder.default)
        resp = self.endpoint.post(url, rule_json)
        return RuleResponse(resp, self.endpoint)

    def query(self, object_type):
        """Begin to build a query on the given object type.

        :param object_type: Either :class:`~tempoiq.protocol.device.Device`
                            or :class:`~tempoiq.protocol.sensor.Sensor`
        :rtype: :class:`~tempoiq.protocol.query.builder.QueryBuilder`"""

        return QueryBuilder(self, object_type)

    def read(self, query):
        url = urlparse.urljoin(self.endpoint.base_url, 'read/')
        j = json.dumps(query, default=self.read_encoder.default)
        accept_headers = [self.ERROR_ACCEPT_TYPE, self.DATAPOINT_ACCEPT_TYPE]
        content_header = self.QUERY_CONTENT_TYPE
        headers = media_types(accept_headers, content_header)
        resp = self.endpoint.get(url, j, headers=headers)
        fetcher = make_fetcher(self.endpoint, url, headers)
        if self.read_version == 'v2':
            return SensorPointsResponse(resp, self.endpoint, fetcher)
        else:
            return StreamResponse(resp, self.endpoint, fetcher)

    def search_devices(self, query):
        #TODO - actually use the size param
        url = urlparse.urljoin(self.endpoint.base_url, 'devices/')
        j = json.dumps(query, default=self.read_encoder.default)
        accept_headers = [self.ERROR_ACCEPT_TYPE, self.DEVICE_ACCEPT_TYPE]
        content_header = self.QUERY_CONTENT_TYPE
        headers = media_types(accept_headers, content_header)
        fetcher = make_fetcher(self.endpoint, url, headers)
        return DeviceResponse(self.endpoint.get(url, j, headers=headers),
                              self.endpoint, fetcher)

    def single(self, query):
        url = urlparse.urljoin(self.endpoint.base_url, 'single/')
        j = json.dumps(query, default=self.read_encoder.default)
        fetcher = make_fetcher(self.endpoint, url)
        return SensorPointsResponse(self.endpoint.get(url, j), self.endpoint,
                                    fetcher)

    def write(self, write_request):
        """Write data points to one or more devices and sensors.

        The write_request argument is a dict which maps device keys to device
        data.

        The device data is itself a dict mapping sensor key to a list of
        :class:`tempoiq.protocol.point.Point`

        :param dict write_request:"""

        url = urlparse.urljoin(self.endpoint.base_url, 'write/')
        default = self.write_encoder.default
        resp = self.endpoint.post(url, json.dumps(write_request,
                                                  default=default))
        return Response(resp, self.endpoint)



from query.selection import ScalarSelectable


#used for enabling the search DSL for substream selection, not up for release
#yet
class Stream(object):
    function = ScalarSelectable('streams', 'function')

    def __init__(self, device, sensor, function):
        self.device = device
        self.sensor = sensor
        self.function = function



import warnings
import exceptions
from selection import Selection, ScalarSelector, OrClause, AndClause
from selection import Compound, DictSelectable
from functions import *
from tempoiq.protocol import Rule
from tempoiq.tempo_exceptions import TempoIQDeprecationWarning


PIPEMSG = 'Pipeline functions passed to monitor call currently have no effect'
DEVICEMSG = 'Pipeline functions passed to device reads have no effect'
ROLLUPMSG = 'Rollup, find, and multi-rollup must have a start and end passed to them'
DELETEMSG = 'Deleting data from sensors requires a start and end time'
DELETEKEYMSG = 'Deleting data from a sensor requires a selection specifying one device key and one sensor key only'
DELETEDEVICEMSG = 'Start and end are invalid arguments for deleting devices.  Are you sure you didn\'t mean session.query(Sensor).delete() instead?'
LATESTMSG = 'The latest() method has been deprecated. Please use single("latest") instead.'


def extract_key_for_monitoring(selection):
    if hasattr(selection.selection, 'selectors'):
        if len(selection.selection.selectors) > 1:
            msg = 'monitoring rules may only be read out by one key at a time'
            raise ValueError(msg)
        return selection.selection.selectors[0].value
    else:
        return selection.selection.value


class QueryBuilder(object):
    """Class to build queries. All instance methods in this class return
    the instance itself. This allows you to construct queries fluently,
    by chaining method calls."""

    def __init__(self, client, object_type):
        self.client = client
        self.object_type = object_type.__name__.lower() + 's'
        self.selection = {
            'devices': Selection(),
            'sensors': Selection(),
            'rules': Selection()
        }
        self.ordering = None
        self.pipeline = []
        self.operation = None

    def _handle_monitor_read(self, **kwargs):
        key = extract_key_for_monitoring(self.selection['rules'])
        method_name = kwargs['__method$$']
        method = getattr(self.client.monitoring_client, method_name)
        return method(key)

    def _normalize_pipeline_functions(self, start, end):
        for function in self.pipeline:
            if isinstance(function, (Rollup, MultiRollup, Find)):
                if start is None or end is None:
                    raise ValueError(ROLLUPMSG)
                if function.args[-1] is None:
                    function.args[-1] = start
            elif isinstance(function, Interpolation):
                if function.args[-2] is None:
                    function.args[-2] = start
                if function.args[-1] is None:
                    function.args[-1] = end

    def _validate_datapoint_delete(self):
        if issubclass(self.selection['devices'].selection.__class__,
                      (Compound, DictSelectable)):
            raise ValueError(DELETEKEYMSG)
        if issubclass(self.selection['sensors'].selection.__class__,
                      (Compound, DictSelectable)):
            raise ValueError(DELETEKEYMSG)
        if self.selection['devices'].selection is None:
            raise ValueError(DELETEKEYMSG)
        if self.selection['sensors'].selection is None:
            raise ValueError(DELETEKEYMSG)
        return (self.selection['devices'].selection.value,
                self.selection['sensors'].selection.value)

    def aggregate(self, function):
        """Aggregate the data in the query with the specified aggregation
        function"""
        self.pipeline.append(Aggregation(function))
        return self

    def annotations(self):
        if not isinstance(self.object_type, Rule):
            raise TypeError('Annotations only applies to monitoring rules')
        key = extract_key_for_monitoring(self.selection['rules'])
        return self.client.monitoring_client.get_annotations(key)

    def changes(self):
        if not isinstance(self.object_type, Rule):
            raise TypeError('Changes only applies to monitoring rules')
        key = extract_key_for_monitoring(self.selection['rules'])
        return self.client.monitoring_client.get_changelog(key)

    def convert_timezone(self, tz):
        """Convert the result's data points to the specified time zone.

        :param String tz: Time zone"""
        self.pipeline.append(ConvertTZ(tz))
        return self

    def delete(self, **kwargs):
        """Execute an API call to delete the objects that are a result of this
        query. Currently only supported for deleting entire devices.
        equivalent to passing this QueryBuilder to
        :meth:`tempoiq.client.Client.delete_device`"""
        if self.object_type == 'devices':
            start = kwargs.get('start')
            end = kwargs.get('end')
            if start or end:
                raise ValueError(DELETEDEVICEMSG)

            self.operation = APIOperation('find', {'quantifier': 'all'})
            return self.client.delete_device(self)
        elif self.object_type == 'sensors':
            start = kwargs.get('start')
            end = kwargs.get('end')
            if start is None or end is None:
                raise ValueError(DELETEMSG)
            (device_key, sensor_key) = self._validate_datapoint_delete()
            args = {'start': start, 'stop': end, 'device_key': device_key,
                    'sensor_key': sensor_key}
            self.operation = APIOperation('delete', args)
            r = self.client.delete_from_sensors(device_key, sensor_key,
                                                start, end)
            return r
        elif self.object_type == 'rules':
            key = extract_key_for_monitoring(self.selection['rules'])
            return self.client.monitoring_client.delete_rule(key)

    def filter(self, selector):
        """Filter the query based on the provided selector. The argument may be
        a :class:`~tempoiq.protocol.query.selection.ScalarSelector` or the
        result of combining several selectors
        with :func:`~tempoiq.protocol.query.selection.or_` or
        :func:`~tempoiq.protocol.query.selection.and_`\ .

        :param selector:
        """
        if not isinstance(selector, (ScalarSelector, OrClause, AndClause)):
            raise TypeError('Invalid object for filter: "%s"' % selector)
        self.selection[selector.selection_type].add(selector)
        return self

    def find(self, function, period, start=None):
        self.pipeline.append(Find(function, period, start))
        return self

    def interpolate(self, function, period, start=None, end=None):
        """Interpolate the sensor data

        :param String function: Interpolation function ("zoh" or "linear")
        :param String period: Time period to interpolate"""
        self.pipeline.append(Interpolation(function, period))
        return self

    def logs(self):
        if not isinstance(self.object_type, Rule):
            raise TypeError('Logs only applies to monitoring rules')
        key = extract_key_for_monitoring(self.selection['rules'])
        return self.client.monitoring_client.get_logs(key)

    def monitor(self, rule):
        if self.pipeline:
            warnings.warn(PIPEMSG, exceptions.FutureWarning)
        rule.selection = self.selection
        return self.client.monitor(rule)

    def multi_rollup(self, functions, period, start=None):
        """Apply multiple rollups to the same sensor data.

        :param list functions: list of rollup functions to apply
        :param String period: Time period of the rollups
        """
        self.pipeline.append(MultiRollup(functions, period, start))
        return self

    def order_by(self, attr, direction):
        self.ordering = {
            'attribute': attr,
            'direction': direction
        }
        return self

    def rollup(self, function, period, start=None):
        """Apply a rollup function to the query.

        :param String function: The rollup function to apply
        :param String period: The time period of the rollup
        :param DateTime start: (optional) A timestamp to use as the start of
        the first rollup period. Default is the same as the start of the read.
        """
        self.pipeline.append(Rollup(function, period, start))
        return self

    def read(self, **kwargs):
        """Make the API call to the TempoIQ backend for this query.

        :param start: required when reading sensor data. Start of time range
                      to read.
        :type start: DateTime
        :param end: required when reading sensor data. End of time range to
                    read.
        :type end: DateTime
        """
        if self.object_type == 'sensors':
            start = kwargs['start']
            end = kwargs['end']
            limit = kwargs.get('limit')
            args = {'start': start, 'stop': end}
            if limit is not None:
                args['limit'] = limit
            #this is set here to be used by the encoder to correctly specify
            #the last step of the operation in the JSON
            self.operation = APIOperation('read', args)
            self._normalize_pipeline_functions(start, end)
            return self.client.read(self)
        elif self.object_type == 'devices':
            if self.pipeline:
                self.pipeline = []
                warnings.warn(DEVICEMSG, exceptions.FutureWarning)
            self.operation = APIOperation('find',
                                          {'quantifier': 'all'})
            return self.client.search_devices(self)
        elif self.object_type == 'rules':
            kwargs['__method$$'] = 'get_rule'
            return self._handle_monitor_read(**kwargs)
        else:
            msg = 'Only sensors, devices, and rules can be selected'
            raise TypeError(msg)

    def single(self, function, timestamp=None, include_selection=False):
        """Make a single-point API call to the TempoIQ backend for this query.

        :param String function: Method for finding the point to return for
                    each sensor. Ex: earliest, latest, before, after
        :param DateTime timestamp: required for all functions except earliest
        """
        if self.object_type == 'sensors':
            args = {'include_selection': include_selection,
                    'function': function}
            if timestamp is not None:
                args['timestamp'] = timestamp

            self.operation = APIOperation('single', args)
            return(self.client.single(self))
        else:
            msg = 'Single value only applies to sensors'
            raise TypeError(msg)

    def latest(self, include_selection=False):
        """Deprecated. Use the
        :func:`~tempoiq.protocol.query.builder.QueryBuilder.single` call
        instead."""
        warnings.warn(LATESTMSG, TempoIQDeprecationWarning)
        self.single('latest', include_selection=include_selection)

    def usage(self):
        if not isinstance(self.object_type, Rule):
            raise TypeError('Usage only applies to monitoring rules')
        key = extract_key_for_monitoring(self.selection['rules'])
        return self.client.monitoring_client.get_usage(key)



from rule import Rule, Condition, Trigger, Filter, Webhook
from device import Device
from sensor import Sensor
from query.selection import Selection, ScalarSelector, AndClause, OrClause


def decode_attributes_selector(selector, selection_type='devices'):
    return ScalarSelector(selection_type, 'attributes', selector['attributes'])


def decode_key_selector(selector, selection_type='devices'):
    return ScalarSelector(selection_type, 'key', selector['key'])


def decode_scalar_selector(selector, selection_type='devices'):
    key = selector.keys()[0]
    if key == 'attributes':
        return decode_attributes_selector(selector, selection_type)
    else:
        return decode_key_selector(selector, selection_type)


def decode_compound_clause(selectors, selection_type='devices', t='and'):
    if t == 'and':
        clause = AndClause()
    else:
        clause = OrClause()

    for s in selectors:
        subt = s.keys()[0]
        if subt in ['and', 'or']:
            subclause = decode_compound_clause(s[subt], selection_type, subt)
            clause.add(subclause)
        else:
            parsed = decode_scalar_selector(s, selection_type)
            clause.add(parsed)

    return clause


def decode_selection(selection, selection_type='devices'):
    s = Selection()
    for k in selection:
        if k in ['and', 'or']:
            selector = decode_compound_clause(selection[k], selection_type, k)
        else:
            selector = decode_scalar_selector(selection, selection_type)
        s.add(selector)
    return s


class DeviceDecoder(object):
    def __init__(self):
        pass

    def __call__(self, dct):
        if dct.get('sensors'):  # Main device object
            return self.decode_device(dct)
        return dct

    def decode_device(self, dct):
        key = dct['key']
        attributes = dct['attributes']
        name = dct['name']
        sensors = []
        for s in dct['sensors']:
            _sensor = Sensor(s['key'], name=s['name'], attributes=s['attributes'])
            sensors.append(_sensor)

        return Device(key, name=name, attributes=attributes, sensors=sensors)


class TempoIQDecoder(object):
    def __init__(self):
        pass

    def __call__(self, dct):
        return self.decode(dct)

    def decode(self, dct):
        if dct.get('rule'):
            return self.decode_rule(dct)
        return dct

    def decode_rule(self, rule):
        name = rule['rule']['name']
        key = rule['rule']['key']
        alert_by = rule['alerts']
        action = Webhook(rule['rule']['action']['url'])

        conditions = []
        for c in rule['rule']['conditions']:
            trigger = Trigger(c['trigger']['name'],
                              c['trigger']['arguments'])
            filters = []
            for f in c['filter']['and']:
                _filter = Filter(f['operation'], f['type'], f['arguments'])
                filters.append(_filter)
            condition = Condition(filters, trigger)
            conditions.append(condition)

        selection = {
            'devices': decode_selection(rule['search']['filters']['devices'],
                                        selection_type='devices'),
            'sensors': decode_selection(rule['search']['filters']['sensors'],
                                        selection_type='sensors'),
        }

        return Rule(name, alert_by=alert_by, key=key, conditions=conditions,
                    action=action, selection=selection)

    def decode_selection(self, dct):
        pass



import re
import dateutil.parser
import pytz


ISO = re.compile(
    r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}'
    '((.\d{3,6})?([+-](\d{4|\d{2}:\d{2}}))?)?')


def check_time_param(t):
    """Check whether a string sent in matches the ISO8601 format.  If a
    Datetime object is passed instead, it will be converted into an ISO8601
    compliant string.

    :param t: the datetime to check
    :type t: string or Datetime
    :rtype: string"""

    if type(t) is str:
        if not ISO.match(t):
            raise ValueError('Date string "%s" does not match ISO8601 format' %
                             (t))
        return t
    else:
        return t.isoformat()


def convert_iso_stamp(t, tz=None):
    """Convert a string in ISO8601 form into a Datetime object.  This is mainly
    used for converting timestamps sent from the TempoDB API, which are
    assumed to be correct.

    :param string t: the timestamp to convert
    :rtype: Datetime object"""

    if t is None:
        return None

    dt = dateutil.parser.parse(t)
    if tz is not None:
        timezone = pytz.timezone(tz)
        if dt.tzinfo is None:
            dt = timezone.localize(dt)
    return dt



import uuid
from tempoiq.temporal.validate import convert_iso_stamp
from device import Device
from sensor import Sensor
from stream import Stream
from point import Point
from query.selection import AndClause, Compound, OrClause, ScalarSelector, and_
from query.selection import Selection


class NoResultError(Exception):
    pass


class TooManyResultsError(Exception):
    pass


class Row(object):
    """Data from one or more sensors at a single timestamp. Returned when
    reading sensor data.

    Example values dict of a row with a single sensor, *temperature*\ , on a
    single device, *test1*\ ::

        {'test1': {'temperature': 500.0} }

    :var timestamp: DateTime of the sensor data
    :var values: dict mapping device key to a dict of sensor keys to values
    """
    def __init__(self, row_json):
        self.timestamp = convert_iso_stamp(row_json['t'])
        self.values = row_json['data']

    def __getitem__(self, key):
        return self.values[key]

    def __iter__(self):
        for device in self.values:
            for sensor in self.values[device]:
                yield ((device, sensor), self.values[device][sensor])


class PointStream(object):
    def __init__(self, stream_info, manager):
        self.stream_info = stream_info
        self.manager = manager
        self._id = str(stream_info['id'])
        self.key = str(uuid.uuid1())
        self._device = None
        self._sensor = None
        self.function = stream_info.get('function')

    def __iter__(self):
        while True:
            data = self.manager.next(self)
            time_str = data['t']
            value = data['data'].get(self._id)
            if value is None:
                continue
            timestamp = convert_iso_stamp(time_str)
            yield Point(timestamp, value)

    @property
    def device(self):
        if self._device is None:
            si = self.stream_info['device']
            self._device = Device(si['key'], si.get('name', ''),
                                  si['attributes'])
        return self._device

    @property
    def sensor(self):
        if self._sensor is None:
            si = self.stream_info['sensor']
            self._sensor = Sensor(si['key'], si.get('name', ''),
                                  si['attributes'])
        return self._sensor


class StreamInfo(object):
    def __init__(self, headers):
        self.headers = headers

    def filter(self, selection):
        evaluator = SelectionEvaluator(selection)
        return evaluator.filter(self.headers)

    def get_one(self, **kwargs):
        selection = self._compile_kwargs(kwargs)
        evaluator = SelectionEvaluator(selection)
        results = [r for r in evaluator.filter(self.headers)]
        if len(results) < 1:
            raise NoResultError('Selection would return no results')
        elif len(results) > 1:
            msg = 'Selection would return more than one result'
            raise TooManyResultsError(msg)
        else:
            return results[0]

    def _compile_kwargs(self, kwargs):
        selectors = []
        for k in kwargs:
            if k == 'device_key':
                selectors.append(Device.key == kwargs[k])
            elif k == 'sensor_key':
                selectors.append(Sensor.key == kwargs[k])
            elif k == 'device_name':
                selectors.append(Device.name == kwargs[k])
            elif k == 'sensor_name':
                selectors.append(Sensor.name == kwargs[k])
            elif k == 'device_attributes':
                for ki in kwargs[k]:
                    selectors.append(Device.attributes[ki] == kwargs[k][ki])
            elif k == 'sensor_attributes':
                for ki in kwargs[k]:
                    selectors.append(Sensor.attributes[ki] == kwargs[k][ki])
            elif k == 'function':
                selectors.append(Stream.function == kwargs[k])
            else:
                msg = 'Invalid bind argument given: "%s"' % k
                raise ValueError(msg)
        selection = Selection()
        selection.add(and_(selectors))
        return selection


class SelectionEvaluator(object):
    def __init__(self, selection):
        self.selection = selection

    def filter(self, headers):
        for header in headers:
            result = self._evaluate_selector(self.selection.selection, header)
            if result:
                yield header

    def _evaluate_compound_clause(self, clause, header):
        if isinstance(clause, AndClause):
            return self._evaluate_and_clause(clause, header)
        elif isinstance(clause, OrClause):
            return self._evaluate_or_clause(clause, header)
        else:
            raise ValueError('Invalid compound clause in selection')

    def _evaluate_and_clause(self, clause, header):
        for selector in clause.selectors:
            matches = self._evaluate_selector(selector, header)
            if not matches:
                return False
        return True

    def _evaluate_or_clause(self, clause, header):
        all_matches = []
        for selector in clause.selectors:
            matches = self._evaluate_selector(selector, header)
            all_matches.append(matches)
        if any(all_matches):
            return True
        return False

    def _evaluate_selector_on_object(self, selector, header, object_type):
        if selector.key == 'key':
            return selector.value == header[object_type]['key']
        elif selector.key == 'name':
            return selector.value == header[object_type]['name']
        elif selector.key == 'function':
            return selector.value == header.get('function')
        else:
            key = selector.value.keys()[0]
            header_value = header[object_type]['attributes'].get(key)
            return selector.value[key] == header_value

    def _evaluate_device_selector(self, selector, header):
        return self._evaluate_selector_on_object(selector, header, 'device')

    def _evaluate_sensor_selector(self, selector, header):
        return self._evaluate_selector_on_object(selector, header, 'sensor')

    def _evaluate_stream_selector(self, selector, header):
        return self._evaluate_selector_on_object(selector, header, 'stream')

    def _evaluate_scalar_selector(self, selector, header):
        if selector.selection_type == 'devices':
            return self._evaluate_device_selector(selector, header)
        elif selector.selection_type == 'sensors':
            return self._evaluate_sensor_selector(selector, header)
        elif selector.selection_type == 'streams':
            return self._evaluate_stream_selector(selector, header)
        else:
            raise ValueError('Invalid selection type in selection')

    def _evaluate_selector(self, selector, header):
        if isinstance(selector, Compound):
            return self._evaluate_compound_clause(selector, header)
        elif isinstance(selector, ScalarSelector):
            return self._evaluate_scalar_selector(selector, header)
        else:
            raise ValueError('Invalid selector in selection')



import requests
from requests.auth import HTTPBasicAuth
import urlparse
import urllib


def make_url_args(params):
    """Utility function for constructing a URL query string from a dictionary
    of parameters.  The dictionary's values can be of various types: lists,
    tuples, dictionaries, strings, or None.

    :param dict params: the key-value pairs to construct a query string from
    :rtype: string"""

    p = []
    for key, value in params.iteritems():
        if isinstance(value, (list, tuple)):
            for v in value:
                p.append((key, v))
        elif isinstance(value, dict):
            for k, v in value.items():
                p.append(('%s[%s]' % (key, k), v))
        elif isinstance(value, bool):
            p.append((key, str(value).lower()))
        elif value is None:
            continue
        else:
            p.append((key, str(value)))
    return urllib.urlencode(p).encode("UTF-8")


def merge_headers(h1, h2):
    return dict(h1.items() + h2.items())


def media_types(accept, content):
    return {
        'Accept': ','.join(accept),
        'Content-Type': content
    }


def media_type(media_resource, media_version, suffix='json'):
    return 'application/prs.tempoiq.%s.%s+%s' % (
        media_resource, media_version, suffix)


class HTTPEndpoint(object):
    """Represents an HTTP endpoint for accessing a REST API.  Provides
    utility methods for GET, POST, PUT, and DELETE requests. Do not explicitly
    call the methods on this class, use the :class:`tempoiq.client.Client`
    class instead.

    :param string base_url: the base URL for the endpoint
    :param string key: the API key for the endpoint
    :param string secret: the API secret for the endpoint"""

    def __init__(self, base_url, key, secret):
        if base_url.endswith('/'):
            self.base_url = base_url + 'v2/'
        else:
            #in case people use their own, it really has to end in a
            #slash so the urljoins will work properly
            self.base_url = base_url + '/v2/'

        self.headers = {
            'User-Agent': 'tempoiq-python/%s' % "1.0.2",
            'Accept-Encoding': 'gzip'
        }
        self.auth = HTTPBasicAuth(key, secret)
        self.pool = requests.session()
        for p in ['http://', 'https://']:
            adapter = requests.adapters.HTTPAdapter()
            self.pool.mount(p, adapter)

    def post(self, url, body, headers={}):
        """Perform a POST request to the given resource with the given
        body.  The "url" argument will be joined to the base URL this
        object was initialized with.

        :param string url: the URL resource to hit
        :param string body: the POST body for the request
        :rtype: requests.Response object"""

        to_hit = urlparse.urljoin(self.base_url, url)
        merged = merge_headers(self.headers, headers)
        resp = self.pool.post(to_hit, data=body, auth=self.auth,
                              headers=merged)
        return resp

    def get(self, url, body='', headers={}):
        """Perform a GET request to the given resource with the given URL.  The
        "url" argument will be joined to the base URL this object was
        initialized with.

        :param string url: the URL resource to hit
        :rtype: requests.Response object"""

        to_hit = urlparse.urljoin(self.base_url, url)
        merged = merge_headers(self.headers, headers)
        resp = self.pool.get(to_hit, data=body, auth=self.auth,
                             headers=merged)
        return resp

    def delete(self, url, body='', headers={}):
        """Perform a DELETE request to the given resource with the given.  The
        "url" argument will be joined to the base URL this object was
        initialized with.

        :param string url: the URL resource to hit
        :rtype: requests.Response object"""

        to_hit = urlparse.urljoin(self.base_url, url)
        merged = merge_headers(self.headers, headers)
        resp = self.pool.delete(to_hit, data=body, auth=self.auth,
                                headers=merged)
        return resp

    def put(self, url, body, headers={}):
        """Perform a PUT request to the given resource with the given
        body.  The "url" argument will be joined to the base URL this
        object was initialized with.

        :param string url: the URL resource to hit
        :param string body: the PUT body for the request
        :rtype: requests.Response object"""

        to_hit = urlparse.urljoin(self.base_url, url)
        merged = merge_headers(self.headers, headers)
        resp = self.pool.put(to_hit, data=body, auth=self.auth,
                             headers=merged)
        return resp



import json
from query.selection import AndClause, Compound, OrClause, ScalarSelector


class TempoIQEncoder(json.JSONEncoder):
    def encode_point(self, point):
        return {
            't': self.encode_datetime(point.timestamp),
            'v': point.value}

    def encode_datetime(self, dt):
        return dt.isoformat()


class WriteEncoder(TempoIQEncoder):
    encoders = {
        'Device': 'encode_device',
        'Sensor': 'encode_sensor',
        'Point': 'encode_point',
        'datetime': 'encode_datetime',
        'Rule': 'encode_rule',
        'Trigger': 'encode_trigger',
        'Webhook': 'encode_webhook'
    }

    def default(self, o):
        encoder_name = self.encoders.get(o.__class__.__name__)
        if encoder_name is None:
            super(TempoIQEncoder, self).default(o)
        encoder = getattr(self, encoder_name)
        return encoder(o)

    def encode_condition(self, condition):
        return {
            'trigger': self.encode_trigger(condition.trigger),
            'filter': {
                'and': map(self.encode_filter, condition.filters)
            }
        }

    def encode_device(self, device):
        return device.key

    def encode_filter(self, _filter):
        return {
            'operation': _filter.inclusion,
            'type': _filter.filter_type,
            'arguments': _filter.args
        }

    def encode_rule(self, rule):
        read_encoder = ReadEncoder()
        j = {
            'conditions': map(self.encode_condition, rule.conditions),
            'name': rule.name,
            'alerts': rule.alert_by,
            'actions': [self.default(rule.action)],
            'selection': {
                'search': {
                    'filters': {
                        'devices': read_encoder.default(
                            rule.selection['devices']),
                        'sensors': read_encoder.default(
                            rule.selection['sensors'])
                    }
                }
            }
        }

        if rule.key is not None:
            j['key'] = rule.key
        return j

    def encode_sensor(self, sensor):
        return sensor.key

    def encode_trigger(self, trigger):
        return {
            'name': trigger.trigger_type,
            'arguments': trigger.args
        }

    def encode_webhook(self, webhook):
        return {
            'url': webhook.url
        }


class CreateEncoder(TempoIQEncoder):
    encoders = {
        'Device': 'encode_device',
        'Sensor': 'encode_sensor'
    }

    def default(self, o):
        encoder_name = self.encoders.get(o.__class__.__name__)
        if encoder_name is None:
            super(TempoIQEncoder, self).default(o)
        encoder = getattr(self, encoder_name)
        return encoder(o)

    def encode_device(self, device):
        return {
            'key': device.key,
            'name': device.name,
            'attributes': device.attributes,
            'sensors': map(self.encode_sensor, device.sensors)
        }

    def encode_sensor(self, sensor):
        return {
            'key': sensor.key,
            'name': sensor.name,
            'attributes': sensor.attributes
        }


class ReadEncoder(TempoIQEncoder):
    encoders = {
        'Point': 'encode_point',
        'datetime': 'encode_datetime',
        'ScalarSelector': 'encode_scalar_selector',
        'AndClause': 'encode_compound_clause',
        'OrClause': 'encode_compound_clause',
        'QueryBuilder': 'encode_query_builder',
        'Selection': 'encode_selection',
        'Find': 'encode_function',
        'Interpolation': 'encode_function',
        'MultiRollup': 'encode_function',
        'Rollup': 'encode_function',
        'Aggregation': 'encode_function',
        'ConvertTZ': 'encode_function'
    }

    def default(self, o):
        encoder_name = self.encoders.get(o.__class__.__name__)
        if encoder_name is None:
            super(TempoIQEncoder, self).default(o)
        encoder = getattr(self, encoder_name)
        return encoder(o)

    def encode_compound_clause(self, clause):
        name = None
        if isinstance(clause, AndClause):
            name = 'and'
        else:
            name = 'or'

        object_type = None
        result = []
        for selector in clause.selectors:
            if isinstance(selector, (AndClause, OrClause)):
                result.append(self.encode_compound_clause(selector))
            elif isinstance(selector, ScalarSelector):
                if object_type is None:
                    object_type = selector.selection_type
                else:
                    if object_type != selector.selection_type:
                        msg = 'Selectors in a compound clause must be of uniform type'
                        raise TypeError(msg)
                result.append(self.encode_scalar_selector(selector))
            else:
                raise ValueError("invalid selector type")

        return {
            name: result
        }

    def encode_function(self, function):
        return {
            'name': function.name,
            'arguments': function.args
        }

    def encode_query_builder(self, builder):
        j = {
            'search': {
                'select': builder.object_type,
                'filters': {
                    'devices': self.encode_selection(
                        builder.selection['devices']),
                    'sensors': self.encode_selection(
                        builder.selection['sensors'])
                }
            },
            builder.operation.name: builder.operation.args
        }

        if builder.ordering is not None:
            j['search']['ordering'] = builder.ordering

        if not j['search']['filters']['devices']:
            if not j['search']['filters']['sensors']:
                j['search']['filters']['devices'] = 'all'
                j['search']['filters']['sensors'] = 'all'
            else:
                j['search']['filters']['devices'] = 'all'
        else:
            if not j['search']['filters']['sensors']:
                j['search']['filters']['sensors'] = 'all'

        if len(builder.pipeline) > 0:
            j['fold'] = {
                'functions': map(self.encode_function, builder.pipeline)

            }
        return j

    def encode_scalar_selector(self, selector):
        return {
            selector.key: selector.value
        }

    def encode_selection(self, selection):
        if selection.selection is None:
            return {}
        if isinstance(selection.selection, Compound):
            if len(selection.selection.selectors) == 0:
                return {}
            else:
                return self.default(selection.selection)
        return self.encode_scalar_selector(selection.selection)



#!/usr/bin/env python
# encoding: utf-8
"""
tempoiq-python/setup.py

Copyright (c) 2012-2015 TempoDB Inc. All rights reserved.
"""

from setuptools import setup


install_requires = [
    'python-dateutil',
    'requests>=1.0',
    'simplejson',
    'pytz',
    'sphinx'
]

tests_require = [
    'mock',
    'unittest2',
]

setup(
    name="tempoiq",
    version="1.0.1",
    author="TempoIQ Inc",
    author_email="aaron.brenzel@tempoiq.com",
    url="http://github.com/tempoiq/tempoiq-python/",
    description="Python bindings for the TempoIQ API",
    packages=["tempoiq", "tempoiq.temporal", "tempoiq.protocol",
              "tempoiq.protocol.query"],
    long_description="Python bindings for the TempoIQ API",
    dependency_links=[
    ],
    setup_requires=['nose>=1.0'],
    install_requires=install_requires,
    tests_require=tests_require,
)



class Function(object):
    def __init__(self, name, args):
        self.name = name
        self.args = args


class Aggregation(Function):
    def __init__(self, function):
        super(Aggregation, self).__init__('aggregation', [function])


class ConvertTZ(Function):
    def __init__(self, tz):
        super(ConvertTZ, self).__init__('convert_tz', [tz])


class Rollup(Function):
    def __init__(self, function, period, start=None):
        super(Rollup, self).__init__('rollup', [function, period, start])


class MultiRollup(Function):
    def __init__(self, functions, period, start=None):
        super(MultiRollup, self).__init__('multi_rollup', [functions, period, start])


class Find(Function):
    def __init__(self, function, period, start=None):
        super(Find, self).__init__('find', [function, period, start])


class Interpolation(Function):
    def __init__(self, function, period, start=None, end=None):
        super(Interpolation, self).__init__('interpolate', [function, period, start, end])


class APIOperation(object):
    def __init__(self, name, args):
        self.name = name
        self.args = args



from client import Client
from endpoint import HTTPEndpoint


def get_session(url, key, secret, read_version='v2'):
    """Get a :class:`tempoiq.client.Client` instance with the given session
    information.

    :param String url: Backend's base URL, in the form
                       "https://your-url.backend.tempoiq.com"
    :param String key: API key
    :param String secret: API secret
    :rtype: :class:`tempoiq.client.Client`"""
    endpoint = HTTPEndpoint(url, key, secret)
    return Client(endpoint, read_version=read_version)
